--[[--------------------------------------------------------------------------------------------------
AudioHandler

Version: 		004
Author: 		Torben Ratzlaff
Website: 		screenheadmedia.com

Purpose:
Makes it easy to use sounds and music and optimizes their performance.
----------------------------------------------------------------------------------------------------]]


--[[--------------------------------------------------------------------------------------------------
How to use:

Use AudioHandler.newAudioData() to create a new set of audio data. You can specify different sounds or streamsin a single audio data. They will share the same properties.

Use AudioHandler.newAudio() to create a specific sound or stream from a preloaded audio data.


--creates a new set of audio data
AudioHandler.newAudioData(name, { 	
	audioType 		= "sound",	--defines the audio type of all audios in this set and determines how these are loaded. Can either be "sound" or "stream".
	fileType 		= "mp3",	--defines the file type of all files in this set e.g. "mp3" or "wav".
	soundNames 		= {"name"}	--a table with all the filenames of the audio that should be included in this set. Defaults so the sets name.
	volume 			= 1,		--set a global volume for all audio in this set.
	maxVolume		= 1,		--any volume set to sound effects of this audio data is mulitplied by this value
	loops			= 0,		--determines how often audios should loop until they are stopped. 0 - play once, 1 - play twice, -1 - loop infinitly.
	duration 		= nil,		--sets the duration for all audios in this set. Defaults to the audios total length.
	fadeIn 			= 0,		--lets all audios of this set fade in for the given amount of milliseconds. 
	maxCopies		= nil,		--determines how many copies of audios of this set should exist at the same time. There is no limitation by default.
	override		= false,	--if the maxCopies of this audio are exceeded, should the oldest one be overridden by newer ones?
	startChannel 	= 1,		--determines at which channel the handler should search for a free channel.
})
--name 		- [required] name of the audio data


--loads several audio data sets from a given data table
AudioHandler.loadAudioDataSet({
	{								--table of a single audio data to load
		name 			= nil 		--[required] name of the audio data
		audioType 		= "sound",	--defines the audio type of all audios in this set and determines how these are loaded. Can either be "sound" or "stream".
		fileType 		= "mp3",	--defines the file type of all files in this set e.g. "mp3" or "wav".
		soundNames 		= {"name"}	--a table with all the filenames of the audio that should be included in this set. Defaults so the sets name.
		volume 			= 1,		--set a global volume for all audio in this set.
		maxVolume		= 1,		--any volume set to sound effects of this audio data is mulitplied by this value
		loops			= 0,		--determines how often audios should loop until they are stopped. 0 - play once, 1 - play twice, -1 - loop infinitly.
		duration 		= nil,		--sets the duration for all audios in this set. Defaults to the audios total length.
		fadeIn 			= 0,		--lets all audios of this set fade in for the given amount of milliseconds. 
		maxCopies		= nil,		--determines how many copies of audios of this set should exist at the same time. There is no limitation by default.
		override		= false,	--if the maxCopies of this audio are exceeded, should the oldest one be overridden by newer ones?
		startChannel 	= 1,		--determines at which channel the handler should search for a free channel.
	},
})


--stops and deletes all playing audios from this set. Deletes the stored data as well.
AudioHandler.deleteAudioData(name) 	
--name 		- [required] name of the audio data


--stops an unloads several audio data
AudioHandler.unloadAudioDataSet({ 	-- table with audio data names
	"name1",	- name of the audio data
	"name2",	- name of the audio data
})


--stops all playing audios from this set.
AudioHandler.clearAudioData(name)	
--name 		- [required] name of the audio data


--creates and plays a new soundEffect from preloaded data
AudioHandler.newAudio(name, {
	soundName 		= "name",	--name of the specific sound in the audio data you want to play. Defaults to the set name.
	volume 			= nil,		--set the volume for the sound. Defaults to the sets volume property.
	override 		= nil,		--if the maxCopies of this audio are exceeded, should the oldest one be overridden by newer ones? Defaults to the sets override property.
	loops			= 0,		--determines how often the audio should loop until it stopps. 0 - play once, 1 - play twice, -1 - loop infinitly.
	duration 		= nil,		--sets the duration for the audio. Defaults to the audios total length.
	fadeIn 			= 0,		--lets the audio fade in for the given amount of milliseconds.
	callback 		= nil 		--a callback function that is called when the sound effects stopped
})
--name 		- [required] name of the audio data
--return	- the new soundEffect


--stops an deletes a specific soundEffect
AudioHandler.deleteAudio(soundEffect, name)
--soundEffect 	- the soundEffect that should be deleted
--name 			- [optional] name of the audio data


--preloads the audio data from a loaded audio data
AudioHandler.preloadAudio(name) 
--name 		- name of the audio data


--pauses a specific soundEffect
AudioHandler.pauseAudio(soundEffect)
--soundEffect 	- the soundEffect that should be paused


--resumes a specific soundEffect
AudioHandler.resumeAudio(soundEffect)
--soundEffect 	- the soundEffect that should be resumed


--rewinds a specific soundEffect
AudioHandler.rewindAudio(soundEffect)
--soundEffect 	- the soundEffect that should be rewind


--set the volume of a given soundEffect
AudioHandler.setVolume(soundEffect, volume)
--soundEffect 	- the soundEffect whos volume you want to set


--fade the soundEffect to a given volume
AudioHandler.fadeAudio(soundEffect, {
	volume 		= nil,	--the volume to fade to
	time 		= nil,	--the amount of time the fade should happen in
	callback 	= nil,	--a callback function that is called when the fade finished
})
--soundEffect 	- the soundEffect you want to fade


--cancels the fading of a soundEffect
AudioHandler.cancelFading(soundEffect, status)
--soundEffect 	- the soundEffect whos fading you want to cancel
--status 		- the status the soundEffect should have after the cancelling


--pauses all sound effects
AudioHandler.pauseAll(flag)
--flag 			- only pauses sound effects with a soecific flag (set in AudioHandler.newAudio())


--resumes all sound effects
AudioHandler.resumeAll(flag)
--flag 			- only pauses sound effects with a soecific flag (set in AudioHandler.newAudio())


--rewinds all sound efects
AudioHandler.rewindAll(flag)
--flag 			- only pauses sound effects with a soecific flag (set in AudioHandler.newAudio())


--fade all sound effects
AudioHandler.fadeAll(arg, flag)
--arg 			- the fading arguments. See AudioHandler.fade() for further information.
--flag 			- only pauses sound effects with a soecific flag (set in AudioHandler.newAudio())


--sets the volume of all sound effects
AudioHandler.setVolumeAll(volume, flag)
--volume 		- the volume you want to set the sound effect to. See AudioHandler.setVolume() for further information.
--flag 			- only pauses sound effects with a soecific flag (set in AudioHandler.newAudio())


----------------------------------------------------------------------------------------------------]]

local setmetatable = setmetatable
local table_indexOf = table.indexOf
local table_remove = table.remove
local type = type
local pairs = pairs
local audio_stop = audio.stop
local audio_dispose = audio.dispose
local audio_pause = audio.pause
local audio_play = audio.play
local audio_resume = audio.resume
local audio_findFreeChannel = audio.findFreeChannel
local audio_setVolume = audio.setVolume
local audio_getVolume = audio.getVolume
local audio_isChannelActive = audio.isChannelActive
local audio_isChannelPaused = audio.isChannelPaused
local audio_loadStream = audio.loadStream
local audio_loadSound = audio.loadSound
local audio_rewind = audio.rewind

local fps =  display.fps

local emptyTable = {}

local AudioHandler --forward referencing


--[[--------------------------------------------------------------------------------------------------
-- setup SoundEffect class
----------------------------------------------------------------------------------------------------]]
local SoundEffect = {}
SoundEffect.__index = SoundEffect

SoundEffect.maxVolume 	= 1
SoundEffect.volume 		= 1

local SoundEffectInstances = {}

function SoundEffect:onComplete(event)
	if self.callback then
		self.callback(self)
	end

	AudioHandler.deleteAudio(self, name)
end


--[[--------------------------------------------------------------------------------------------------
-- setup AudioHandler Module
----------------------------------------------------------------------------------------------------]]

local audioBasePath = "audio/"

local allAudioData = {} 	--contains all audio data and all loaded audios
local allFadingAudio = {} 	--contains all fading audios


AudioHandler = {}


function AudioHandler.newAudioData(name, arg)
	if not allAudioData[name] then
		local arg = arg or emptyTable

		local audioData = {
			activeAudio = {},
			arguments 	= {
				audioType 		= arg.audioType 	or "sound",
				fileType 		= arg.fileType 		or "mp3",
				soundNames 		= arg.soundNames 	or emptyTable,
				volume 			= arg.volume 		or 1,
				maxVolume		= arg.maxVolume 	or 1,
				loops			= arg.loops 		or 0,
				duration 		= arg.duration,
				fadeIn 			= arg.fadeIn 		or 0,
				maxCopies		= arg.maxCopies,
				override		= arg.override 		or false,
				startChannel 	= arg.startChannel 	or 1,
			},
		}
		
		allAudioData[name] = audioData
	else
		print("AudioHandler - It already exists an effect with the name: "..name)
	end
end


function AudioHandler.loadAudioDataSet(dataSet)
	for i=1, #dataSet do
		local audioData = dataSet[i]

		AudioHandler.newAudioData(audioData.name, audioData)
	end
end


function AudioHandler.deleteAudioData(name)
	local audioData = allAudioData[name]
	if audioData then
		for i=#SoundEffectInstances, 1, -1 do
			local soundEffect = SoundEffectInstances[i]

			if soundEffect.name == name then
				audio_stop(soundEffect.channel)

				soundEffect.status = "deleted"

				table_remove(SoundEffectInstances, table_indexOf(SoundEffectInstances, soundEffect))
			end
		end

		audioData.activeAudio = {}
		
		if audioData.loaded then
			audio_dispose(audioData.loaded)
			audioData.loaded = nil
		end
	else
		print("AudioHandler - There exists no effect with the name: "..name)
	end
end


function AudioHandler.unloadAudioDataSet(dataNames)
	for i=1, #dataNames do
		AudioHandler.deleteAudioData(dataNames[i])
	end
end


function AudioHandler.clearAudioData(name)
	local audioData = allAudioData[name]
	if audioData then
		for i=#SoundEffectInstances, 1, -1 do
			local soundEffect = SoundEffectInstances[i]

			if soundEffect.name == name then
				audio_stop(soundEffect.channel)

				soundEffect.status = "deleted"

				table_remove(SoundEffectInstances, table_indexOf(SoundEffectInstances, soundEffect))
			end
		end

		audioData.activeAudio = {}
	else
		print("AudioHandler - There exists no effect with the name: "..name)
	end
end


function AudioHandler.newAudio(name, arg)
	local audioData = allAudioData[name]
	
	if audioData then
		local activeAudio 		= audioData.activeAudio
		local arguments 		= audioData.arguments
		local loaded 			= audioData.loaded

		local currentArguments = arg or emptyTable
	
		local soundName = currentArguments.soundName or arguments.soundNames[1] or name

		local maxCopies 		= arguments.maxCopies
		local audioType 		= arguments.audioType
		local startChannelNum 	= arguments.startChannel
		local lastChannel 		= arguments.lastChannel

		local volume 			= currentArguments.volume 	or arguments.volume
		local maxVolume			= arguments.maxVolume
		local override 			= currentArguments.override or arguments.override
		
		volume = volume*maxVolume

		--checks if audio has already been loaded
		if not loaded then
			loaded = AudioHandler.preloadAudio(name)
		end

		local loadedSound = loaded[soundName]

		local channel
		
		--search for a free channel.
		if maxCopies and #activeAudio > maxCopies and audioType ~= "stream" then
			--cancels the sound that is playing for the longest time
			if override == true then
				channel = table_remove(activeAudio, 1)
				audio_stop(channel)
			else
				print("AudioHandler - There are already playing "..#activeAudio.." copies of the effect with the name: "..name)
				return
			end
		else
			channel = audio_findFreeChannel(startChannelNum)
			
			if audioType == "stream" and channel == lastChannel then
				channel = audio_findFreeChannel(channel+1)
			end
		end
		
		--starts new audio if a channel is found
		if channel >= startChannelNum and channel <= 32 then
			local soundEffect = setmetatable({}, SoundEffect)

			local actualChannel = audio_play(loadedSound, {
				channel 	= channel,
				loops 		= currentArguments.loops 	or arguments.loops,
				duration 	= currentArguments.duration or arguments.duration,
				onComplete 	= function(event) soundEffect:onComplete(event) end,
			})
			
			--check if the audio is really playing (bug)
			if actualChannel ~= 0 then
				SoundEffectInstances[#SoundEffectInstances + 1] = soundEffect

				soundEffect.name 		= name
				soundEffect.channel 	= actualChannel
				soundEffect.audioType 	= audioType
				soundEffect.maxVolume	= maxVolume
				soundEffect.status 		= "playing"
				soundEffect.callback 	= currentArguments.onComplete
				soundEffect.flag		= currentArguments.flag
				
				if audioType == "stream" then
					arguments.lastChannel = actualChannel
				end
				
				activeAudio[#activeAudio + 1] = actualChannel
				

				local fadeIn = currentArguments.fadeIn or arguments.fadeIn

				if fadeIn and fadeIn > 0 then
					AudioHandler.setVolume(soundEffect, 0)
					AudioHandler.fadeAudio(soundEffect, {time = fadeIn, volume = volume})
				else
					AudioHandler.setVolume(soundEffect, volume)
				end
					
				return soundEffect
			else
				arguments.lastChannel = nil
			
				print("AudioHandler - Couldn't find free channel for effect with the name: "..name)
			end
		else
			arguments.lastChannel = nil
		
			print("AudioHandler - Couldn't find free channel for effect with the name: "..name)
		end
	else
		print("AudioHandler - There exists no effect with the name: "..name)
	end
end


function AudioHandler.deleteAudio(soundEffect, name)
	if soundEffect then
		local name = name or soundEffect.name
		local audioData = allAudioData[name]
		
		if audioData and soundEffect.status ~= "deleted" then
			local channel = soundEffect.channel
			
			local activeAudio = audioData.activeAudio
			
			local index = table_indexOf(activeAudio, channel)
			

			if soundEffect.status == "fading" then
				AudioHandler.cancelFading(soundEffect)
			end

			soundEffect.status = "deleted"
			
			if index then
				table_remove(activeAudio, index)
				
				audio_stop(channel)
			end

			table_remove(SoundEffectInstances, table_indexOf(SoundEffectInstances, soundEffect))
		else
			print("AudioHandler - There exists no effect with the name: "..name)
		end
	else
		print("AudioHandler - Need audio handle to delete audio")
	end
end	


function AudioHandler.preloadAudio(name)
	local audioData = allAudioData[name]
	
	if audioData then
		local arguments = audioData.arguments
		local audioType = arguments.audioType
		local fileType 	= arguments.fileType

		local arguments_soundNames = arguments.soundNames
		
		local soundNames
		if arguments_soundNames and #arguments_soundNames > 0 then
			soundNames = arguments_soundNames
		else
			soundNames = {name}
		end
		
		if not audioData.loaded then
			local loadedAudio = {}
			audioData.loaded = loadedAudio
			
			if audioType == "stream" then
				for i=1, #soundNames do
					local soundName = soundNames[i]
					loadedAudio[soundName] = audio_loadStream(audioBasePath..soundName.."."..fileType)
				end
			else
				for i=1, #soundNames do
					local soundName = soundNames[i]
					loadedAudio[soundName] = audio_loadSound(audioBasePath..soundName.."."..fileType)
				end
			end

			return loadedAudio
		else
			print("AudioHandler - There is already an loaded soundfile for the effect with the name: "..name)
		end
	else
		print("AudioHandler - There exists no effect with the name: "..name)
	end
end


function AudioHandler.pauseAudio(soundEffect)
	if soundEffect then
		local channel = soundEffect.channel
		
		audio_pause(channel)
		
		soundEffect.status = "paused"
	else
		print("AudioHandler - Need audio handle to pause audio")
	end
end


function AudioHandler.resumeAudio(soundEffect)
	if soundEffect then
		local channel = soundEffect.channel
		
		audio_resume(channel)
		
		soundEffect.status = "playing"
	else
		print("AudioHandler - Need audio handle to resume audio")
	end
end


function AudioHandler.rewindAudio(soundEffect)
	if soundEffect then
		local audioType = soundEffect.audioType
		
		if audioType == "stream" then
			local name = soundEffect.name
			local audioData = allAudioData[name]
			local audioHandle = audioData.loaded[name]
			
			audio_rewind(audioHandle)
		else
			local channel = soundEffect.channel
		
			audio_rewind(channel)
		end
	else
		print("AudioHandler - Need audio handle to rewind audio")
	end
end


function AudioHandler.setVolume(soundEffect, volume)
	if soundEffect then
		local channel = soundEffect.channel
		local name = soundEffect.name
	
		local audioData = allAudioData[name]
		local audioArguments = audioData.arguments
		
		local volume = volume or audioArguments.volume
		volume = volume*soundEffect.maxVolume
		
		audio_setVolume(volume, {channel = channel})
		soundEffect.currentVolume = volume
	else
		print("AudioHandler - Need audio handle to set volume")
	end
end


function AudioHandler.fadeAudio(soundEffect, arg)
	if soundEffect then
		local arg = arg or emptyTable
		local time 		= arg.time
		local volume 	= arg.volume
		
		local channel = soundEffect.channel
		
		local currentVolume = audio_getVolume({channel = channel})

		
		if time and volume and time > 0 and volume ~= currentVolume then
			if soundEffect.status == "fading" then
				AudioHandler.cancelFading(soundEffect)
			end
			
			soundEffect.status = "fading"
			
			local valueDifference 	= volume*soundEffect.maxVolume - currentVolume
			local frames 			= time * 0.001 * fps
			local valueChange 		= valueDifference / frames

			soundEffect.fadingArguments = 
			{
				endValue 	= volume,
				valueChange = valueChange,
				callback 	= arg.callback,
			}
			
			allFadingAudio[#allFadingAudio + 1] = soundEffect
		end
	end
end


function AudioHandler.cancelFading(soundEffect, status)
	if soundEffect then
		if soundEffect.status == "fading" then
			local new_status
			
			if status then
				soundEffect.status = status
			else
				local channel = soundEffect.channel
				local channelActive = audio_isChannelActive(channel)
				local channelPaused = audio_isChannelPaused(channel)
				
				if channelActive then
					if channelPaused then
						soundEffect.status = "paused"
					else
						soundEffect.status = "playing"
					end
				else
					soundEffect.status = "deleted"
				end
			end
			
			table_remove(allFadingAudio, table_indexOf(allFadingAudio, soundEffect))
		end
	end
end


function AudioHandler.pauseAll(flag)
	for i=1, #SoundEffectInstances do
		local soundEffect = SoundEffectInstances[i]

		if flag == soundEffect.flag then
			AudioHandler.pauseAudio(soundEffect)
		end
	end
end


function AudioHandler.resumeAll(flag)
	for i=1, #SoundEffectInstances do
		local soundEffect = SoundEffectInstances[i]

		if flag == soundEffect.flag then
			AudioHandler.resumeAudio(soundEffect)
		end
	end
end


function AudioHandler.rewindAll(flag)
	for i=1, #SoundEffectInstances do
		local soundEffect = SoundEffectInstances[i]

		if flag == soundEffect.flag then
			AudioHandler.rewindAudio(soundEffect)
		end
	end
end


function AudioHandler.fadeAll(arg, flag)
	for i=1, #SoundEffectInstances do
		local soundEffect = SoundEffectInstances[i]

		if flag == soundEffect.flag then
			AudioHandler.fadeAudio(soundEffect, arg)
		end
	end
end


function AudioHandler.setVolumeAll(volume, flag)
	for i=1, #SoundEffectInstances do
		local soundEffect = SoundEffectInstances[i]

		if not flag or flag == soundEffect.flag then
			AudioHandler.setVolume(soundEffect, volume)
		end
	end
end


--handles all the fading
function AudioHandler_enterFrame()
	for i = #allFadingAudio, 1, -1 do
		local soundEffect = allFadingAudio[i]
		local fadingArguments = soundEffect.fadingArguments
		
		if soundEffect.status ~= "deleted" and fadingArguments then
			local channel 	= soundEffect.channel
			local name 		= soundEffect.name
			
			local endValue 		= fadingArguments.endValue
			local valueChange 	= fadingArguments.valueChange
			local callback 		= fadingArguments.callback
			
			local newVolume 	= soundEffect.currentVolume + valueChange
			
			local kicked
			
			if (valueChange > 0 and newVolume >= endValue) or (valueChange < 0 and newVolume <= endValue) then
				newVolume =  endValue
				
				kicked = true
			end
			
			if newVolume < 0 then newVolume = 0
			elseif newVolume > 1 then newVolume = 1 end
			
			audio_setVolume(newVolume, {channel = channel})
			soundEffect.currentVolume = newVolume

			if kicked then
				table_remove(allFadingAudio, table_indexOf(soundEffect))
				
				if callback then
					local callbackType = type(callback)
					if callbackType == "string" then
						AudioHandler[callback](soundEffect)
					elseif callbackType == "table" then
						callback[1](callback[2])
					else
						callback()
					end
				end
			end
		else
			table_remove(allFadingAudio, table_indexOf(soundEffect))
		end
	end
end

Runtime:addEventListener("enterFrame", AudioHandler_enterFrame)


return AudioHandler